from copy import deepcopy
from urllib.parse import urlparse

from pb_common.config.configure import ConfigCreator
from pb_common.db import AbstractDbProvider
from pb_common.db.ad_provider import LdapProvider
from pb_common.db.oracle_provider import OracleProvider
from pb_common.db.redis_provider import RedisProvider
from pb_common.services.dbo_provider import Dbo2ServiceProvider
from pb_common.services.fimi_provider import FimiProvider
from pb_common.utils.logger import get_logger


logger = get_logger(__name__)


class UnknownDbException(Exception):
    pass


class UnknownServiceException(Exception):
    pass


class UnknownConfigException(Exception):
    pass


class BaseContext(object):
    def __init__(self, config):
        self._env = ConfigCreator(config).reload()
        self._services = dict()
        self._db_registry = dict()
        self._path = []
        logger.debug('Context {} is running!'.format(self))

    @property
    def config(self):
        return self._env

    def _databases_close(self):
        for name, database in self._db_registry.items():
            if isinstance(database, AbstractDbProvider):
                database.close()

    def close(self):
        """
        Закрыть соединения с БД
        :return:
        """
        self._databases_close()
        logger.debug("Shutdown context!")

    def _get_item_in_config(self, *path, name):
        """
        Получить объект из конфига
        :param path: путь до объекта
        :param name: ключ объекта
        :return: объект конфига
        """
        env = self._env
        for item in path:
            env = getattr(env, item, None)
            if not env:
                raise UnknownConfigException('Не найден конфиг: {}'.format(item))
        return getattr(env, name, None)

    def get_service(self, *path, name, provider, **kwargs):
        """
        Получить провайдер для работы с сервисом
        :param path: путь до сервиса в конфиге
        :param name: имя сервиса
        :param provider: провайдер для работы с сервисом
        :return: инстанс провайдера для работы с сервисом
        """
        if name not in self._services:
            env = self._get_item_in_config(*path, name=name)
            if not env:
                raise UnknownServiceException('Не найден конфиг сервиса: {}'.format(name))
            if kwargs:
                env.update(kwargs)
            self._services.update({name: provider(service='{} Service'.format(name.title()), **dict(env))})
        return self._services[name]

    def get_db_provider(self, *path, name, provider, query_storage=None):
        """
        Получить провайдер для работы с БД
        :param path: путь до БД в конфиге
        :param name: имя БД
        :param provider: провайдер для работы с БД
        :param query_storage: хранилище запросов к БД
        :return: объект для работы с БД
        """
        if name not in self._db_registry:
            env = self._get_item_in_config(*path, name=name)
            if not env:
                raise UnknownDbException('Не найден конфиг БД: {}'.format(name))
            self._db_registry.update({name: provider(**dict(env)).connect()})
            if query_storage:
                self._db_registry[name].query_storage = query_storage
        return self._db_registry[name]


class Context(BaseContext):
    @property
    def client_service(self):
        """
        Провайдер для запросов в ClientService
        """
        return self.get_service('services', 'dbo', name='client', provider=Dbo2ServiceProvider)

    @property
    def card_service(self):
        """
        Провайдер для запросов в CardService
        """
        return self.get_service('services', 'dbo', name='card', provider=Dbo2ServiceProvider)

    @property
    def session_service(self):
        """
        Провайдер для запросов в SessionService
        """
        return self.get_service('services', 'dbo', name='session', provider=Dbo2ServiceProvider)

    @property
    def registration_service(self):
        """
        Провайдер для запросов в RegistrationService
        """
        ib_url = self.config.get('baseUrls', {}).get('ib')
        if ib_url:
            url = urlparse(ib_url)
            if url.port:
                return self.get_service(
                    'services', 'dbo', name='registration', provider=Dbo2ServiceProvider, port=url.port)
        return self.get_service('services', 'dbo', name='registration', provider=Dbo2ServiceProvider)

    @property
    def contract_service(self):
        """
        Провайдер для запросов в ContractService
        """
        return self.get_service('services', 'dbo', name='contract', provider=Dbo2ServiceProvider)

    @property
    def token_provider_service(self):
        """
        Провайдер для запросов в TokenProvider
        """
        return self.get_service('services', 'dbo', name='tokenprovider', provider=Dbo2ServiceProvider)

    @property
    def bender_hub(self):
        """
        Провайдер для запросов в Bender (получение тестовых пользователей)
        """
        return self.get_service('services', name='benderhub', provider=Dbo2ServiceProvider)

    @property
    def fimi(self):
        """
        Провайдер для запросов в Мультикарту
        """
        return self.get_service('services', 'esb', name='fimi', provider=FimiProvider)

    @property
    def redis_provider(self):
        """
        Провайдер для запросов в Redis
        """
        if not self._db_registry.get('redis'):
            env = self._env.db.redis
            env.hosts = list(map(lambda x: x if isinstance(x, tuple) else tuple(x.split(':')), env.hosts))
            self._db_registry.update({'redis': RedisProvider(**dict(env)).connect()})
        return self._db_registry['redis']

    @property
    def ldap_provider(self):
        """
        Провайдер для запросов в LDAP (хранится авторизацонные данные клиентов)
        """
        return self.get_db_provider('db', name='ldap', provider=LdapProvider)

    @property
    def auth_provider(self):
        """
        Провайдер для работы с БД Oracle -  AUTH (хранится авторизацонные данные клиентов)
        """
        return self.get_db_provider('db', name='auth', provider=OracleProvider)

    @property
    def vabank_provider(self):
        """
        Провайдер для работы с БД Oracle -  VABANK (договоры клиентов)
        """
        return self.get_db_provider('db', name='vabank', provider=OracleProvider)

    @property
    def siebel_provider(self):
        """
        Провайдер для работы с БД Oracle - Siebel (информация о клиенте, заявках и договорах)
        """
        return self.get_db_provider('db', name='siebel', provider=OracleProvider)

    @property
    def cdr_provider(self):
        """
        Провайдер для работы с БД Oracle - CDR
        """
        return self.get_db_provider('db', name='cdr', provider=OracleProvider)

    @property
    def sas_provider(self):
        """
        Провайдер для работы с БД Oracle - SAS
        """
        return self.get_db_provider('db', name='sas', provider=OracleProvider)
